unit crtsock;

{
  CrtSocket for Delphi 32
  Copyright (C) 1999  Paul Toth <tothpaul@free.fr>
  http://tothpaul.free.fr

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

}

{
 16 march 2002 - add MultiCast support

}
interface

uses windows,sysutils;

{-$define debug}

// Server side :
//  - start a server
//  - wait for a client
function StartServer(Port:word):integer;
function WaitClient(Server:integer):integer;
function WaitClientEx(Server:integer; var ip:string):integer;

// Client side :
//  - call a server
function CallServer(Server:string;Port:word):integer;

// Both side :
//  - Assign CRT Sockets
//  - Disconnect server
procedure AssignCrtSock(Socket:integer; Var Input,Output:TextFile);
procedure Disconnect(Socket:integer);

// Udp
function SendText(Server:string; Port:word; Text:string):integer;

// BroadCasting (UDP)
function StartBroadCast(Port:word):integer;
function SendBroadCast(Handle:integer; Port:word; s:string):integer;
function SendBroadCastTo(Handle:integer; Port:word; ip,s:string):integer;
function ReadBroadCast(Handle:integer):string;
function ReadBroadCastEx(Handle:integer; var ip:string):string;

// Multicast (UDP)
function JoinMultiCast(address:string; port:word; ttl:integer):integer;
procedure LeaveMulticast(socket:integer; address:string);

// BlockRead
function SockAvail(Socket:integer):integer;
function DataAvail(Var F:TextFile):integer;
Function BlockReadsock(Var F:TextFile; var s:string):boolean;

function SendData(socket:integer; data:pointer; datalen:integer):integer;
function ReadData(socket:integer; data:pointer; datalen:integer):integer;

Function send(socket:integer; data:pointer; datalen,flags:integer):integer; stdcall; far;
Function recv(socket:integer; data:pchar; datalen,flags:integer):integer; stdcall; far;

// some usefull SOCKET apis
type
  PHost = ^THost;
  THost = packed record
    name     : PChar;
    aliases  : ^PChar;
    addrtype : Smallint;
    length   : Smallint;
    addr     : ^pointer;
  end;

  TSockAddr=packed record
   Family:word;
   Port:word;
   Addr:LongInt;
   Zeros:array[0..7] of byte;
  end;

  TTimeOut=packed record
   sec:integer;
   usec:integer;
  end;

Const
 fIoNbRead = $4004667F;
 SD_RECEIVE     = 0;
 SD_SEND        = 1;
 SD_BOTH        = 2;

Function socket(Family,Kind,Protocol:integer):integer; stdcall;
Function closesocket(socket:Integer):integer; stdcall;
Function gethostbyname(HostName:PChar):PHost; stdcall;
Function gethostname(name:pchar; size:integer):integer; stdcall;
Function bind(Socket:Integer; Var SockAddr:TSockAddr; AddrLen:integer):integer; stdcall;
Function WSAGetLastError:integer; stdcall;
Function ioctlsocket(socket:integer; cmd: integer; var arg: integer): Integer; stdcall;
Function shutdown(Socket,How:Integer):integer; stdcall;
Function recvfrom(socket:integer; data:pointer; datalen,flags:integer; var SockAddr:TSockAddr; var AddrLen:Integer):integer; stdcall; 

// Convert an IP Value to xxx.xxx.xxx.xxx string
Function LongToIp(Long:LongInt):string;
Function IpToLong(ip:string):longint;
Function HostToLong(AHost:string):LongInt;
Function LocalHost:string;

Var
 EofSock:boolean;

implementation

//------ winsock -------------------------------------------------------
Const
 WinSock='wsock32.dll'; { 32bits socket DLL }
 Internet=2; { Internet familly }
 Stream=1;   { Streamed socket }
 Datagrams=2;
// fIoNbRead = $4004667F;
 IPPROTO_IP=0;
 SOL_SOCKET=$FFFF;
 SO_REUSEADDR    = $0004;
 SO_BROADCAST    = $0020;          { permit sending of broadcast msgs }

 IP_MULTICAST_TTL  =3; // or 10 for Winsock 2 !
 IP_MULTICAST_LOOP =4; // or 11 for Winsock 2 !
 IP_ADD_MEMBERSHIP =5; // or 12 for Winsock 2 !
 IP_DROP_MEMBERSHIP=6; // or 13 for Winsock 2 !


Type
  TWSAData = packed record
    wVersion: Word;
    wHighVersion: Word;
    szDescription: array[0..256] of Char;
    szSystemStatus: array[0..128] of Char;
    iMaxSockets: Word;
    iMaxUdpDg: Word;
    lpVendorInfo: PChar;
  end;

{ Winsock }
Function WSAStartup(Version:word; Var Data:TwsaData):integer; stdcall; far; external winsock;
Function socket(Family,Kind,Protocol:integer):integer; stdcall; far; external winsock;
Function shutdown(Socket,How:Integer):integer; stdcall; far; external winsock;
Function closesocket(socket:Integer):integer; stdcall; far; external winsock;
Function WSACleanup:integer; stdcall; far; external winsock;
//Function WSAAsyncSelect(Socket:Integer; Handle:Hwnd; Msg:word; Level:Longint):longint; stdcall; far; external winsock;
Function bind(Socket:Integer; Var SockAddr:TSockAddr; AddrLen:integer):integer; stdcall; far; external winsock;
Function listen(socket,flags:Integer):integer; stdcall; far; external winsock;
Function connect(socket:Integer; Var SockAddr:TSockAddr; AddrLen:integer):integer; stdcall; far; external winsock;
Function accept(socket:Integer; Var SockAddr:TSockAddr; Var AddrLen:Integer):integer; stdcall; far; external winsock;
Function WSAGetLastError:integer; stdcall; far; external winsock;
Function recv(socket:integer; data:pchar; datalen,flags:integer):integer; stdcall; far; external winsock;
Function send(socket:integer; data:pointer; datalen,flags:integer):integer; stdcall; far; external winsock;
//Function getpeername(socket:integer; var SockAddr:TSockAddr; Var AddrLen:Integer):Integer; stdcall; far; external winsock;
Function gethostbyname(HostName:PChar):PHost; stdcall; far; external winsock;
//Function getsockname(socket:integer; var SockAddr:TSockAddr; Var AddrLen:Integer):integer; stdcall; far; external winsock;
//Function inet_ntoa(addr:longint):PChar; stdcall; far; external winsock;
Function WSAIsBlocking:boolean; stdcall; far; external winsock;
Function WSACancelBlockingCall:integer; stdcall; far; external winsock;
Function ioctlsocket(socket:integer; cmd: integer; var arg: integer): Integer; stdcall; far; external winsock;
//Function gethostbyaddr(var addr:longint; size,atype:integer):PHost; stdcall; far; external winsock;
Function gethostname(name:pchar; size:integer):integer; stdcall; far; external winsock;
function select(nfds:integer; readfds, writefds, exceptfds:pointer; var timeout:TTimeOut):integer; stdcall; far; external winsock;
function setsockopt(socket,level,optname:integer;var optval; optlen:integer):integer; stdcall; far; external winsock;
Function sendto(socket:integer; data:pointer; datalen,flags:integer; var SockAddr:TSockAddr; AddrLen:Integer):integer; stdcall; far; external winsock;
Function recvfrom(socket:integer; data:pointer; datalen,flags:integer; var SockAddr:TSockAddr; var AddrLen:Integer):integer; stdcall; far; external winsock;


Function IpToLong(ip:string):LongInt;
 var
  x,i:byte;
  ipx:array[0..3] of byte;
  v:integer;
 begin
  Result:=0;
  longint(ipx):=0; i:=0;
  for x:=1 to length(ip) do
   if ip[x]='.' then begin
    inc(i);
    if i=4 then exit;
   end else begin
    if not (ip[x] in ['0'..'9']) then exit;
    v:=ipx[i]*10+ord(ip[x])-ord('0');
    if v>255 then exit;
    ipx[i]:=v;
   end;
  result:=longint(ipx);
 end;

Function HostToLong(AHost:string):LongInt;
 Var
  Host:PHost;
 begin
  Result:=IpToLong(AHost);
  if Result=0 then begin
   Host:=GetHostByName(PChar(AHost));
   if Host<>nil then Result:=longint(Host^.Addr^^);
  end;
 end;

Function LongToIp(Long:LongInt):string;
 var
  ipx:array[0..3] of byte;
  i:byte;
 begin
  longint(ipx):=long;
  Result:='';
  for i:=0 to 3 do result:=result+IntToStr(ipx[i])+'.';
  SetLength(Result,Length(Result)-1);
 end;

Function LocalHost:string;
 var
  local:array[0..255] of char;
 begin
  if GetHostName(local,sizeof(local))=0 then Result:=local else result:='localhost';
 end;
                      
//--- Server Side ------------------------------------------------------------------------
function StartServer(Port:word):integer;
 Var
  SockAddr:TSockAddr;
 begin
  Result:=socket(Internet,Stream,0);
  if Result=-1 then exit;
  FillChar(SockAddr,SizeOf(SockAddr),0);
  SockAddr.Family:=Internet;
  SockAddr.Port:=swap(Port);
  if (Bind(Result,SockAddr,SizeOf(SockAddr))<>0)
  or (Listen(Result,0)<>0) then begin
   CloseSocket(Result);
   Result:=-1;
  end;
 end;

function WaitClient(Server:integer):integer;
 var
  Client:TSockAddr;
  Size:integer;
 begin
  Size:=SizeOf(Client);
  Result:=Accept(Server,Client,Size);
 end;

function WaitClientEx(Server:integer; var ip:string):integer;
 var
  Client:TSockAddr;
  Size:integer;
 begin
  Size:=SizeOf(Client);
  Result:=Accept(Server,Client,Size);
  ip:=LongToIp(Client.Addr);
 end;

function SockReady(Socket:integer):boolean;
 var
  sockset:packed record
   count:integer;
   socks:{array[0..63] of} integer;
  end;
  timeval:TTimeOut;
 begin
  sockSet.count:=1;
  sockSet.socks:=Socket;
  timeval.sec  :=0;
  timeval.usec :=0;
  result:=Select(0,@sockSet,nil,nil,timeval)>0;
 end;

function SockAvail(Socket:integer):integer;
 var
  rdy:boolean;
 begin
  rdy:=SockReady(Socket); // before IoCtlSocket to be sure (?) that we don't get some data between the 2 calls
  if IoctlSocket(Socket, fIoNbRead,Result)<0 then
   Result:=-1
  else begin
   if (Result=0) and RDY then result:=-1; // SockReady is TRUE when Data ara Avaible AND when Socket is closed
  end;
 end;

function DataAvail(Var F:TextFile):integer;
 var
  s:integer;
 begin
 // cause of TexTFile Buffer, we need to check both Buffer & Socket !
  With TTextRec(F) do begin
   Result:=BufEnd-BufPos;
   s:=SockAvail(Handle);
  end;
  if Result=0 then Result:=s else if s>0 then Inc(Result,s);
 end;

Function BlockReadSock(Var F:TextFile; var s:string):boolean;
 Var
  Handle:THandle;
  Size:integer;
 begin
  Result:=False;
  Handle:=TTextRec(F).Handle;
  Repeat
   if (IoctlSocket(Handle, fIoNbRead, Size)<0) then exit;
   if Size=0 then exit
  until (Size>0);
  SetLength(s,Size);
  Recv(Handle,pchar(s),Size,0);
  Result:=True;
 end;

// Client Side--------------------------------------------------------------------------
function CallServer(Server:string; Port:word):integer;
 var
  SockAddr:TSockAddr;
 begin
  Result:=socket(Internet,Stream,0);
  if Result=-1 then exit;
  FillChar(SockAddr,SizeOf(SockAddr),0);
  SockAddr.Family:=Internet;
  SockAddr.Port:=swap(Port);
  SockAddr.Addr:=HostToLong(Server);
  if Connect(Result,SockAddr,SizeOf(SockAddr))<>0 then begin
   Disconnect(Result);
   Result:=-1;
  end;
 end;

// UDP ---------------------------------------------
function SendText(Server:string; Port:word; Text:string):integer;
 Var
  SockAddr:TSockAddr;
  s:integer;
 begin
  s:=socket(Internet,Stream,0);
  if s=-1 then
   result:=-2
  else begin
   FillChar(SockAddr,SizeOf(SockAddr),0);
   SockAddr.Family:=Datagrams;
   SockAddr.Port:=swap(Port);
   SockAddr.Addr:=HostToLong(Server);
   if Connect(s,SockAddr,SizeOf(SockAddr))<>0 then begin
    Disconnect(s);
    Result:=-3;
    exit;
   end;
   Result:=Send(s,@Text[1],length(Text),0);
   CloseSocket(s);
  end;
 end;

// BroadCasting-------------------------------------
function StartBroadCast(Port:word):integer;
 Var
  SockAddr:TSockAddr;
  bc:integer;
 begin
  Result:=socket(Internet,Datagrams,17); // 17 for UDP ... work also with 0 ?!
  if Result=-1 then exit;
  FillChar(SockAddr,SizeOf(SockAddr),0);
  SockAddr.Family:=Internet;
  SockAddr.Port:=swap(Port);
//  SockAddr.Addr:=0; ?
  bc:=SO_BROADCAST;
  if (Bind(Result,SockAddr,SizeOf(SockAddr))<>0)
  or (setsockopt(Result,SOL_SOCKET,SO_BROADCAST,bc,SizeOf(bc))<>0) then begin
   CloseSocket(Result);
   Result:=-1;
  end;
 end;

function SendBroadCast(Handle:integer; Port:word; s:string):integer;
 Var
  SockAddr:TSockAddr;
 begin
  SockAddr.Family:=Internet;
  SockAddr.Port:=swap(Port);
  SockAddr.Addr:=-1;
  Result:=SendTo(Handle,@s[1],length(s),0,SockAddr,SizeOf(SockAddr));
 end;

function SendBroadCastTo(Handle:integer; Port:word; ip,s:string):integer;
 Var
  SockAddr:TSockAddr;
 begin
  SockAddr.Family:=Internet;
  SockAddr.Port:=swap(Port);
  SockAddr.Addr:=IpToLong(ip);
  Result:=SendTo(Handle,@s[1],length(s),0,SockAddr,SizeOf(SockAddr));
 end;

function ReadBroadCast(Handle:integer):string;
 Var
  SockAddr:TSockAddr;
  SockLen:integer;
  len:integer;
 begin
  FillChar(SockAddr,SizeOf(SockAddr),0);
  SockAddr.Family:=Internet;
  SockLen:=SizeOf(SockAddr);
  setlength(result,1024);
  len:=recvfrom(Handle,@result[1],1024,0,SockAddr,SockLen);
  if len>0 then SetLength(result,len) else result:='';
 end;

function ReadBroadCastEx(Handle:integer; var ip:string):string;
 Var
  SockAddr:TSockAddr;
  SockLen:integer;
  len:integer;
 begin
  FillChar(SockAddr,SizeOf(SockAddr),0);
  SockAddr.Family:=Internet;
  SockLen:=SizeOf(SockAddr);
  setlength(result,1024);
  len:=recvfrom(Handle,@result[1],1024,0,SockAddr,SockLen);
//  writeln('len=',len);
  if len>0 then SetLength(result,len) else result:='';
  ip:=LongToIp(SockAddr.Addr);
 end;

//------------ MultiCast ---------------
function JoinMultiCast(address:string; port:word; ttl:integer):integer;
 Var
  SockAddr:TSockAddr;
  mc:array [0..1] of integer; // multicast
  lb:BOOL;
 begin
  Result:=socket(Internet,Datagrams,0);
  if Result=-1 then exit;
 // reusable socket (multiple instance of the same client on the same machine)
  lb:=TRUE;
  if setsockopt(Result,SOL_SOCKET,SO_REUSEADDR,lb,SizeOf(lb))<>0 then begin
//   writeln('reuseaddr error');
  end;
// bind it
  FillChar(SockAddr,SizeOf(SockAddr),0);
  SockAddr.Family:=Internet;
  SockAddr.Port:=swap(Port);
  mc[0]:=IpToLong(address); // multicast IP group address
  mc[1]:=0; // any ?!
  if (Bind(Result,SockAddr,SizeOf(SockAddr))<>0) then begin
   CloseSocket(Result);
   result:=-2;
  end;
 // join the Multicast ring
  if (setsockopt(Result,IPPROTO_IP,IP_ADD_MEMBERSHIP,mc,SizeOf(mc))<>0) then begin
   CloseSocket(Result);
   result:=-3;
  end;
 // set Multicast TTL
  if setsockopt(Result,IPPROTO_IP,IP_MULTICAST_TTL,ttl,SizeOf(ttl))<>0 then begin
   CloseSocket(Result);
   result:=-4;
  end;
 // avoid Loop back
  lb:=TRUE;
  if setsockopt(Result,IPPROTO_IP,IP_MULTICAST_LOOP,lb,SizeOf(lb))<>0 then begin
//   Writeln('loopback error ',WSAGetLastError);
  end;
 end;

procedure LeaveMulticast(socket:integer; address:string);
 var
  mc:array[0..1] of integer;
 begin
  mc[0]:=IpToLong(address);
  mc[1]:=0;
  setsockopt(socket,IPPROTO_IP,IP_DROP_MEMBERSHIP,mc,SizeOf(mc));
  CloseSocket(socket);
 end;

// read/write

function SendData(socket:integer; data:pointer; datalen:integer):integer;
var
 p:pchar;
 l:integer;
begin
 p:=data;
// l:=datalen;
 result:=0;
 repeat
  l:=send(socket,p,datalen,0);
  if l<=0 then begin
   result:=l;
   exit;
  end;
  inc(result,l);
  inc(p,l);
  dec(datalen,l);
 until datalen=0;
end;

function ReadData(socket:integer; data:pointer; datalen:integer):integer;
var
 p:pchar;
 l:integer;
begin
 p:=data;
// l:=datalen;
 result:=0;
 repeat
  l:=recv(socket,p,datalen,0);
  if l<=0 then begin
   result:=l;
   exit;
  end;
  inc(result,l);
  inc(p,l);
  dec(datalen,l);
 until datalen=0;
end;

//------------ CrtSock -----------------
Var
 InitOk:boolean;

function OutputSock(Var F:TTextRec):integer; far;
 var
  i,l,j:integer;
 begin
 {$ifdef debug}writeln('out ',F.BufPtr);{$endif}
  Result:=0;
  if F.BufPos<>0 then begin
   i:=0;
   l:=F.BufPos;
   repeat
    j:=Send(F.Handle,@F.BufPtr[i],l,0);
    if j<=0 then exit;
    dec(l,j);
    inc(i,j);
   until l=0;
   F.BufPos:=0;
  end;
 end;

function InputSock(var F: TTextRec): Integer; far;
 Var
  Size:integer;
 begin
  F.BufEnd:=0;
  F.BufPos:=0;
  Result:=0;

  Repeat
   if (IoctlSocket(F.Handle, fIoNbRead, Size)<0) then begin
    EofSock:=True;
    exit;
   end;
  until (Size>=0);

  Size:=Recv(F.Handle,F.BufPtr,F.BufSize,0);
  if Size>0 then F.BufEnd:=Size;
  EofSock:=(F.BufEnd<=0);
 {$ifdef debug}writeln('in  ',F.BufPtr);{$endif}
 end;

procedure Disconnect(Socket:integer);
 var
  dummy:array[0..1024] of char;
 begin
  if ShutDown(Socket,1)=0 then
   repeat until recv(Socket,dummy,1024,0)<=0;
  //ShutDown(Socket,2); 
  CloseSocket(Socket);
 end;

function CloseSock(var F:TTextRec):integer; far;
 begin
  Disconnect(F.Handle);
  F.Handle:=-1;
  Result:=0;
 end;

function OpenSock(var F: TTextRec): Integer; far;
begin
  F.BufPos:=0;
  F.BufEnd:=0;
  if F.Mode = fmInput then begin // ReadLn
    EofSock:=False;
    F.InOutFunc := @InputSock;
    F.FlushFunc := nil;
  end else begin                 // WriteLn
    F.Mode := fmOutput;
    F.InOutFunc := @OutputSock;
    F.FlushFunc := @OutputSock;
  end;
  F.CloseFunc := @CloseSock;
  Result:=0;
end;

Procedure AssignCrtSock(Socket:integer; Var Input,Output:TextFile);
 begin
  with TTextRec(Input) do begin
    Handle := Socket;
    Mode := fmClosed;
    BufSize := SizeOf(Buffer);
    BufPtr := @Buffer;
    OpenFunc := @OpenSock;
  end;
  with TTextRec(Output) do begin
    Handle := Socket;
    Mode := fmClosed;
    BufSize := SizeOf(Buffer);
    BufPtr := @Buffer;
    OpenFunc := @OpenSock;
  end;
// Delphi 6 patch ... comment this two lines for Delphi < 6
  SetLineBreakStyle(Input,tlbsCRLF);
  SetLineBreakStyle(Output,tlbsCRLF);

  Reset(Input);
  Rewrite(Output);
 end;

//----- Initialization/Finalization--------------------------------------------------

Procedure InitCrtSock;
 var
  wsaData:TWSAData;
 begin
  InitOk:=wsaStartup($101,wsaData)=0;
{$ifdef debug}allocconsole{$endif}
 end;

Procedure DoneCrtSock;
 begin
  if not InitOk then exit;
  if wsaIsBlocking then wsaCancelBlockingCall;
  wsaCleanup;
 end;

Initialization InitCrtSock;

Finalization DoneCrtSock;

end.
